/* ************************************************************************************** */
/* Program 		: Liquidity_Pontiff_Singla_1.sas                                          */
/* SAS Output	: Final_Liquidity.sas7bdat saved in defined home directory			      */
/* Date Updated	: Jun 2019                                                                */
/* Author  		: Jeffrey Pontiff & Rohit Singla                                          */
/* Paper   		: Pontiff, Jeffrey and Rohit Singla, Liquidity Risk? (Feb 2019). 	      */
/*				  Accepted in Critical Finance Review (CFR) journal					      */
/*			 	  Available at http://dx.doi.org/10.2139/ssrn.3320011					  */
/*			 	  PERMISSION TO USE THIS PROGRAM IS GRANTED UNDER THE CONDITION 		  */
/*			 	  THAT THE USER(S) CITE PONTIFF AND SINGLA'S CFR PUBLICATION	  		  */
/* Summary 		: 1.This program is designed to be executed on the WRDS mainframe, and   */
/*				  replicates Liquidity Factors of Pastor & Stambaugh(PS)                  */
/*          	  originally published in Journal of Political Economy(JPE) in 2003.      */
/*           	  2.The code also produces liquidity estimates for various time frames    */
/*				  using out-of-their-sample data. To obtain the liquidity factors for     */
/*				  other time frames, just change the parameters begyear, daten and 	      */
/*				  endyear on lines 41, 42 and 43, respectively.						      */
/*				  3.To run PS variants in Table 4 of our CFR paper, do the following:	  */ 
/*				  	(a)To remove price restrictions, and including all the prices, assume */
/*					   min_prc=0.001 & max_prc=10000000 on lines 46 and 47, respectively. */				 
/*				    (b)For value weighted index, uncomment lines 329-333 and comment	  */ 
/*					   lines 323-326. Do vice-versa for equal weighted index.			  */
/*				  	(c)To include zero volume, comment line 169, and uncomment line 170.  */
/*					(d)To restrict the intercept to zero, comment line 199, and uncomment */
/*					   line 200.														  */
/* ************************************************************************************** */

*Setting libraries and options;
%include '!sasroot/wrdslib.sas';
libname home '/scratch/bc/pontiff';
options mprint notes source;

*********Major Parameters*******;

*For PS + Post-PS time period (Baseline Scenario), assume begyear = 1961 and start month (daten) = 196208 and endyear = 2017;
*For Pre PS-time period assume begyear = 1926 and start month (daten) = 192701 and endyear = 1967;
*For PS-time period, assume begyear = 1961 and start month (daten) = 196208 and endyear = 1999;
*For Post PS-JPE paper assumes begyear = 1994 and start month (daten) = 199501 and endyear = 2017;
*For entire period replication assumes begyear = 1926 and start month (daten) = 192701 and endyear = 2017;

%let begyear=1961; *beginning year;
%let daten = 196208; *start month;
%let endyear=2018;*end year;

%let dollvol = 0; *Imputed value of dollar volume when missing;
%let min_prc = 5; *Lower price restriction;
%let max_prc = 1000; *Upper price restriction;
%let scale =  300005.7; *Scaling parameter i.e. Market value in August 1962;
%let shrcd = (10, 11); *Share code restriction, sharecode in 10, 11;
%let exchcd = (3, 33, 0); *Exchange code restriction, exchange code not in 3, 33 and 0;


**************************************
STEP 1: Computing Stock Level Liquidity
**************************************

*Form Latest Header Information;

proc sql;
	create table hdr as select c.permco, c.permno, c.cusip, c.hcomnam, c.hexcd, 
		c.hsiccd, c.begret, c.begprc, c.begvol, c.endret, c.endprc, c.endvol from 
		crspa.dsfhdr c order by c.permno;
	update hdr set begvol='01jan25'd where begvol is missing;
	update hdr set endvol='01jan25'd where endvol is missing;
	create table hdr as select permno, permco, hexcd, hsiccd, begret, endret, 
		begvol, endvol from hdr where begret ^=.;
	*Form Daily Dollar Volume;
	create table ddvol as select a.permco, a.permno, a.date, a.vol, a.vol * 
		abs(a.prc) as dvol from crspa.dsf a inner join hdr c on a.permno=c.permno 
		where a.date>=c.begvol and a.date <=c.endvol and year(a.date)>=&begyear and 
		year(a.date) <=&endyear
order by a.permno, a.date;
	create table ddvol as select permno, date, dvol from ddvol;
	*Form Daily Returns;
	create table dret as select a.permno, a.date, case when a.ret=. then -99 when 
		a.ret=.E then -44 when a.ret=.D then -55 when a.ret=.C then -66 when a.ret=.B 
		then -77 when a.ret=.A then -88 else a.ret end as ret from crspa.dsf a inner 
		join crspa.dsfhdr b on a.permno=b.permno where a.date >=b.begret and 
		a.date <=b.endret and year(a.date)>=&begyear and year(a.date) <=&endyear
order by a.permno, a.date;
	*Form Daily Market Returns;
	create table dmret as select date, vwretd, ewretd from crspa.dsi where 
		year(date) between &begyear and &endyear 
order by date;
	*Merging Daily Dollar Volume, Daily Returns and Market Returns;
	create table dmerged as select a.*, b.dvol from dret a inner join ddvol b on 
		a.permno=b.permno and a.date=b.date order by permno, date;
	create table dmerged as select a.*, b.vwretd from dmerged a inner join dmret b 
		on a.date=b.date order by permno, date;
quit;

*Creating Month and Year Attributes from Date;

data dmerged2;
	set dmerged;
	year=year(date);
	month=month(date);
run;

*Creating Lags of Daily Returns, Market Returns and Daily Dollar Volume;

proc expand data=dmerged2 out=dmerged3 method=none;
	id date;
	convert ret=lagret / transformout=(lag 1);
	convert vwretd=lagmkt / transformout=(lag 1);
	convert dvol=lagdvol / transformout=(lag 1);
	by permno;
run;

*Form Trading Calendar;

proc sql;
	create table dcal as select distinct a.date from crspa.dsf a order by a.date;
quit;

*Excluding Stock-Months with First and Last Partial months that it appears on the CRSP tape;

proc expand data=dcal out=dcal2 method=none;
	convert date=lag_date / transformout=(lag 1);
	convert date=lead_date / transformout=(lead 1);
run;

data partial;
	set dmerged3;
	by permno;

	if first.permno or last.permno;
run;

proc sql;
	create table partial2 as select a.*, b.lag_date, b.lead_date from partial a 
		inner join dcal2 b on a.date=b.date order by permno, date;
quit;

data partial3;
	set partial2;
	by permno;

	if first.permno and month(date)=month (lag_date) then
		partial=1;

	if last.permno and month(date)=month (lead_date) then
		partial=1;
run;

proc sql;
	create table dmerged3 as select a.*, b.partial from dmerged3 a left join 
		partial3 b on a.permno=b.permno and a.year=b.year and a.month=b.month order 
		by permno, date;
quit;

data dmerged3;
	set dmerged3;
	where partial ~=1;
	drop partial;
run;

*Creating Excess Returns and signed Dollar Volume variables;

data dmerged4;
	set dmerged3;
	ret_e=ret - vwretd;
	ret_lag_e=lagret - lagmkt;
	dollvol_lag=(sign(ret_lag_e)*lagdvol)/1000000;
run;

data dmerged4;
	set dmerged4;
	where ret>=-1 and lagret>-1 and lagdvol > 0; *This is the way PS do it;
	*where ret>=-1 and lagret>-1 ;  *this allows for zero volume inclusion;
run;

*Counting number of Stocks every month;

proc means data=dmerged4 N noprint;
	by permno year month;
	output out=count N=nobs;
run;

proc sql;
	create table dmerged4 as select a.*, b.nobs from dmerged4 a inner join count b 
		on a.permno=b.permno and a.year=b.year and a.month=b.month;

	*Considering only the months that have > 15 observations. 
	For 09/2011 month the restriction is > 10 observations;

	create table dmerged4 as select * from dmerged4 where (nobs > 15) 
		or (nobs > 10 and year=2001 and month=9);
quit;

proc sort data=dmerged4 out=dmerged5;
	by permno year month date;
run;

*Regressing Excess Returns on Lag of Returns and Lagged Dollar Volume, 
and saving the coefficient of Lagged Dollar Volume (stock-level liquidity measure);

proc reg data=dmerged5 edf outest=est_rohit noprint;
	model ret_e=lagret dollvol_lag; *Normal PS estimation;
	*model ret_e = lagret dollvol_lag  /noint; *saves a degree of freedom;
	by permno year month;
	run;

*Computing number of stocks (n) that went into monthly regressions;

data est_rohit;
	set est_rohit;
	n=_in_ + _edf_ + 1;

	if lagret=0 or dollvol_lag=0 then
		delete;
	keep permno year month n dollvol_lag;
run;

**************************************
STEP 2: Computing Aggregated Liquidity
**************************************;

*Form Latest Header Info;

proc sql;
	create table namdata as select permco, permno, namedt, nameenddt, exchcd, 
		shrcd, siccd from crspa.stocknames;
	create table hdr as select c.permco, c.permno, c.cusip, c.hcomnam, c.hexcd, 
		c.hsiccd, begret, begprc, begvol, endret, endprc, endvol from crspa.msfhdr c 
		order by c.permno;
	update hdr set begvol='01jan25'd where begvol is missing;
	update hdr set endvol='01jan25'd where endvol is missing;
	*Form Monthly Shares, Exchancge and SIC Codes;
	create table mdata as select a.permco, a.permno, a.date, b.exchcd, b.shrcd, 
		b.siccd from crspa.msf a inner join namdata b on a.permno=b.permno and 
		a.date>=b.namedt and a.date<=b.nameenddt inner join hdr c on 
		a.permno=c.permno where a.date>=c.begprc and a.date <=c.endprc order by 
		a.permno, a.date;
	*Form Annnual Shares, Exchancge and SIC Codes;
	create table adata as select permco, permno, year(date) as calyear, exchcd, 
		shrcd, siccd from mdata where month(date)=12 order by permno, date;
quit;

data adatax;
	set adata;
	where calyear >=%eval(&begyear-1) and calyear<=&endyear;
run;

data annexch;
	set adatax;
	keep permno calyear exchcd shrcd;
run;

*Form Monthly Market Value, Monthly Shares Outstanding and Prices;

proc sql;
	create table mval as select a.permno, a.date, case when a.prc=. then 0 when 
		a.shrout=. then 0 else a.shrout * abs(a.prc) end as mval from crspa.msf a 
		inner join hdr c on a.permno=c.permno where a.date>=c.begprc and 
		a.date <=c.endprc order by a.permno, a.date;
	create table mshrout as select a.permno, a.date, a.shrout from crspa.msf a 
		inner join hdr c on a.permno=c.permno where a.date>=c.begprc and 
		a.date <=c.endprc order by permno, a.date;
	create table monprc as select a.*, b.shrout from mval a inner join mshrout b 
		on a.permno=b.permno and a.date=b.date;
quit;

data monprc2;
	set monprc;
	prc=mval/shrout;
	where mval > 0 and shrout > 0;
run;

proc sql;
	create table monvals2 as select a.permno, a.date, a.mval/1000 as mval, 
		abs(b.prc) as prc from mval a left join monprc2 b on a.permno=b.permno and 
		a.date=b.date;
quit;

data monvals2;
	set monvals2;
	year=year(date);
	month=month(date);
run;

*Merging in Monthly Stock_level Liquidity into the Monthly Stock File;

proc sql;
	create table merged as select a.*, b.dollvol_lag from monvals2 a left join 
		est_rohit b on a.permno=b.permno and a.year=b.year and a.month=b.month order 
		by permno, year, month;
	*Merging in the lagged exchange codes and share codes;
	create table merged1 as select a.*, b.exchcd, b.shrcd from merged a left join 
		annexch b on a.permno=b.permno and (a.year-1)=b.calyear order by permno, 
		year, month;
quit;

*Forming lags of Dollar Volume, Prices and Market cap;

proc expand data=merged1 out=merged2 method=none;
	convert dollvol_lag=lagdol /transformout=(lag 1);
	convert prc=lagprc / transformout=(lag 1);
	convert prc=lag2prc / transformout=(lag 2);
	convert mval=lagmval / transformout=(lag 1);
	by permno;
	id date;
run;

*Putting restrictions on lag of prices, share codes and exchange codes;

data merged3;
	set merged2;
	where (&min_prc < lagprc < &max_prc);
run;

data merged4;
	set merged3;
	where shrcd in &shrcd 
	and exchcd not in &exchcd;
run;

* Computing aggregated liquidity and adjusting it with market value in August 1962;

proc sql noprint;

*Equal weighted;
create table final as select year, month, avg(dollvol_lag) as avg_liq, 
	sum(lagmval)/&scale. as adj_mval, sum(lagmval) as totcap, count(dollvol_lag) 
	as totsec from merged4 where dollvol_lag ^=.
group by year, month order by year, month;

*Value weighted;
/*	create table final as select year, month, sum(lagmval)/&scale. as adj_mval, 
		sum(lagmval) as totcap, count(dollvol_lag) as totsec
, avg((dollvol_lag*lagmval))/sum(lagmval) as avg_liq from merged4 where 
		dollvol_lag ^=.
group by year, month order by year, month;*/
quit;

data final2;
	set final;
	agg_liq=avg_liq*adj_mval;
run;

************************************************
STEP 3: Computing innovation in liquidity measure
************************************************;

*Putting restrictions on first and second lags of prices, share codes and exchange codes;

data merged3;
	set merged2;
	where (&min_prc < lagprc < &max_prc) and  (&min_prc < lag2prc < &max_prc);
run;

data merged4;
	set merged3;
	where shrcd in &shrcd and exchcd not in &exchcd;
run;

*Computing change in liquidity;

proc sql;
	create table merged5 as select year, month, avg(dollvol_lag - lagdol) as 
		avgchg_liq from merged4 group by year, month order by year, month;
	*Merging in aggregated liquidity computed in Step 2;
	create table merged5_1 as select a.*, b.agg_liq, b.adj_mval from merged5 a 
		inner join final2 b on a.year=b.year and a.month=b.month;
quit;

data merged5_2;
	set merged5_1;
	avgchg_liq_bf=avgchg_liq;
	avgchg_liq=avgchg_liq*adj_mval;
run;

*Computing lags of change in liquidity and aggregated liquidity;

proc expand data=merged5_2 out=merged6 method=none;
	convert avgchg_liq=lag_avgchg_liq/transformout=(lag 1);
	convert agg_liq=lag_agg_liq/transformout=(lag 1);
run;

*The Following SAS macro (innov) regresses Change in liquidity on its lag and lagged aggregated 
liquidity every year (from 1962 to 2017) till the data available at the end of that year only. 
The residuals of regressions measure the innovations in liquidity measure;

data merged6;
	set merged6;
	daten=year*100+month;
	date=input(put(year*100+month, 6.), yymmn6.);
	format date yymmn6.;
run;

proc datasets lib=work nodetails nolist;
	delete  resid resid2
	;
run;

%let begyear=&begyear+1;

%macro innov;
	%do y=&endyear %to &begyear %by -1;
		%put "now working on year = &y";

		proc reg data=merged6 outest=est_1 noprint;
			where year <=&y 
and daten > &daten;
			m&y.: model avgchg_liq=lag_avgchg_liq lag_agg_liq;

			%if &y=&endyear %then
				%do;
					output out=resid %end;
			%else
				%do;
					output out=resid2 %end;
			r=resid;
		quit;

		%if &y=&endyear %then
			%do;

				data resid;
					retain run;
					set resid;
				run=&y;

				keep run year month resid;
				run;
			%end;
		%else
			%do;

				data resid2;
					retain run;
					set resid2;
				run=&y;

				keep run year month resid;
				run;

				data resid;
					set resid resid2;
					where resid ^=.;
				run;

			%end;

		data dini;
			set merged6;
			where daten=&daten;
		run=&y;

		resid=avgchg_liq;
		keep run year month resid;
		run;

		proc append base=resid data=dini force;
			;
		run;

		proc append base=est_f data=est_1 force;
		run;

	%end;
%mend innov;

%innov 

data est_t;
set est_f;
where lag_avgchg_liq=0 or lag_agg_liq=0;
run;

proc sort data=resid;
	by descending run year month;
run;

*The innovations in liquidity measure at the end of 2017 is saved in innovations datafile;

proc sql;
	create table innovations as select year, month, resid from resid where 
		run=&endyear
order by year, month;
quit;

************************************************************
STEP 4: Computing Traded liquidity (Portfolio Returns spread)
************************************************************;

* Using FF factors directly obtained from WRDS;

data ff2;
	set ff.factors_monthly;
	year=year(date);
	month=month(date);
run;

*Form Monthly Returns;

proc sql;
	create table mret as select a.permno, a.date, case when a.ret=. then -99 when 
		a.ret=.E then -44 when a.ret=.D then -55 when a.ret=.C then -66 when a.ret=.B 
		then -77 when a.ret=.A then -88 else a.ret end as ret from crspa.msf a inner 
		join crspa.msfhdr b on a.permno=b.permno where a.date >=b.begret and 
		a.date <=b.endret order by a.permno, a.date;
quit;

*Replacing dates with years and months;

data mret2;
	set mret;
	year=year(date);
	month=month(date);
run;

*Merging in the lagged exchange codes and share codes;

proc sql;
	create table mret3 as select a.*, b.exchcd, b.shrcd from mret2 a left join 
		annexch b on a.permno=b.permno and (a.year-1)=b.calyear order by permno, 
		year, month;
quit;

*Putting Share codes restrictions only;

data mret4;
	set mret3;
	where shrcd in &shrcd;
run;

*Rescaling Market cap in the market value datafile. Replacing dates with year and months;

data mval2;
	set mval;
	year=year(date);
	month=month(date);
	mval=mval/1000;
run;

*Retaining end-of-the-year prices from monthly prices datafile;

data monprc3;
	set monprc2;
	year=year(date);
	month=month(date);
	where month(date)=12;
	keep permno year prc;
run;

*Putting end-of-the-year price restrictions;

data monprc4;
	set monprc3;
	where &min_prc < prc < &max_prc;
run;

*The following macro regresses Adjusted monthly returns for each stock on innovations (Step 3) 
after controlling for adjusted market returns and Fama French factors (SMB and HML).
The regressions for each stock is run every year based on the monthly data from past 60 months;

%macro sorting;
	%do y=&endyear %to %eval(&begyear+4) %by -1;

		proc sql;
			create table betareg as select a.permno, a.year, a.month
	, a.ret-b.rf as exret
	, b.mktrf as mkt_rf, b.smb, b.hml
	, c.resid from mret4 a inner join ff2 b on a.year=b.year and a.month=b.month 
				inner join resid c on a.year=c.year and a.month=c.month where %eval(&y-5) 
				< a.year <=&y and ret > -1 and run=&y and (a.year*100+a.month) >=&daten
order by permno, year, month;
			create table freq as select permno, count(*) as count from betareg group by 
				permno having count >=60 order by permno;
			create table betareg2 as select a.* from betareg a inner join freq b on 
				a.permno=b.permno;
		quit;

		proc reg data=betareg2 outest=betas noprint;
			model exret=resid mkt_rf smb hml;
			by permno;
		quit;

		data final1;
			retain permno year;
			set betas;
			year=&y;
			rename resid=beta;
			label resid='beta';
			keep permno year resid;

			if resid=0 or mkt_rf=0 or smb=0 or hml=0 then
				delete;	* Remove not full rank regressions;
		run;

		proc sql;
			create table final1_sort as select * from final1 order by beta;

			%if &y=&endyear %then
				%do;
					create table superfinal like final1_sort;
				%end;
			insert into superfinal select * from final1_sort;
		quit;

	%end;
%mend sorting;

%sorting 

data superfinal;
set superfinal;
run;

proc sort data=superfinal out=supf_t1;
	by permno year;
run;

*Keeping only the stocks' betas satisfying end-of-year price restriction ;

proc sql;
	create table superfinal_1 as select a.* from superfinal a inner join monprc4 b 
		on a.permno=b.permno and a.year=b.year;
quit;

*Sorting and ranking betas;

proc sort data=superfinal_1;
	by year beta;
run;

data superfinal_1;
	set superfinal_1;
run;

proc rank data=superfinal_1 out=results ties=mean;
	by year;
	var beta;
	ranks betaRank;
run;

data results;
	set results;
run;

*Constructing top and bottom declile portfolios;

proc sql;
	create table freq2 as select year, count(*) as count from superfinal_1 group 
		by year;
	create table results2 as select a.*, b.count from results a inner join freq2 b 
		on a.year=b.year;
	create table results3 as select permno, year, beta, ceil(10*(betarank/count)) 
		as decile from results2;
	create table top_decile as select * from results3 where decile=1;
	create table bottom_decile as select * from results3 where decile=10;
	*Computing value weighted returns for top and bottom decile portfolios;
	create table retval as select a.*, b.mval from mret2 a inner join mval2 b on 
		a.permno=b.permno and intnx('month', a.date, 0, 'B')=intnx('month', b.date, 
		1, 'B');
	create table top_retval as select a.*, b.year as yearp from retval a inner 
		join top_decile b on a.permno=b.permno and a.year=b.year+1 where mval>0 and 
		ret>-1 order by year, month;
	create table top_retval2 as select permno, year, month, ret, mval, sum(mval) 
		as sum_mval from top_retval group by year, month;
	create table top_retval3 as select year, month, sum(ret*(mval/sum_mval)) as 
		vwret_1 from top_retval2 group by year, month;
	create table bottom_retval as select a.* from retval a inner join 
		bottom_decile b on a.permno=b.permno and a.year=b.year+1 where mval>0 and 
		ret>-1 order by year, month;
	create table bottom_retval2 as select permno, year, month, ret, mval, 
		sum(mval) as sum_mval from bottom_retval group by year, month;
	create table bottom_retval3 as select year, month, sum(ret*(mval/sum_mval)) as 
		vwret_10 from bottom_retval2 group by year, month;
	*Computing 10-1 portfolio returns (or Traded Liquidity);
	create table port10_1 as select a.year, b.month, a.vwret_1, b.vwret_10, 
		(vwret_10 - vwret_1) as traded_liquidity from top_retval3 a inner join 
		bottom_retval3 b on a.year=b.year and a.month=b.month;
	*Merging in Aggregated Liquidity (Step 2), Innovations (Step 3) and Traded Liquidity (Step 4);
	create table home.Final_Liquidity as select a.year, a.month, a.agg_liq 
		label='Our Aggregted Liquidity' , b.resid label='Our Innovations', 
		c.traded_liquidity label='Our Traded Liquidity' from final2 a left join 
		innovations b on a.year=b.year and a.month=b.month left join port10_1 c on 
		a.year=c.year and a.month=c.month;
quit;
run;
